/*  
  Copyright 1997-98 Michael Callahan
  Copyright 1999 Carnegie Mellon University
  This program is free software.  You may copy it according
  to the conditions of the GNU General Public License version 2.0
  or later; see the file COPYING in the source directory.
*/

#include "sockvxd.h"

PCBUF
NewCBuf (int size)
{
	PCBUF buf = 0;

	buf = (PCBUF) _HeapAllocate (sizeof(CBUF), HEAPZEROINIT);
	if (!buf)
		goto error;

	buf->cb_mutex = _CreateMutex (0, 0);
	if (!buf->cb_mutex)
		goto error;

	buf->cb_start = (PCHAR) _HeapAllocate (size, 0);
	if (!buf->cb_start)
		goto error;

	buf->cb_end = buf->cb_start + size;
	buf->cb_head = buf->cb_start;
	buf->cb_tail = buf->cb_start;
	return buf;

error:
	if (buf && buf->cb_start) _HeapFree (buf->cb_start, 0);
	if (buf && buf->cb_mutex) _DestroyMutex (buf->cb_mutex);
	if (buf) _HeapFree(buf, 0);
	return NULL;
}

void
FreeCBuf (PCBUF buf)
{
	_HeapFree (buf->cb_start, 0);
	_DestroyMutex (buf->cb_mutex);
	_HeapFree (buf, 0);
}

int
LenCBuf (PCBUF buf)
{
	if (buf->cb_head >= buf->cb_tail)
		return buf->cb_head - buf->cb_tail;
	return buf->cb_head - buf->cb_start
		 + buf->cb_end - buf->cb_tail;
}

int
AvailCBuf (PCBUF buf)
{
	return buf->cb_end - buf->cb_start - LenCBuf(buf);
}

int
EmptyCBuf (PCBUF buf)
{
	return buf->cb_head == buf->cb_tail;
}

int
ResizeCBuf (PCBUF buf, int newsize)
{
	int len;
	PCHAR nbuf = NULL, obuf = NULL;

	LockCBuf (buf);

	len = LenCBuf(buf);
	if (len >= newsize) 
		goto error;

	nbuf = _HeapAllocate(newsize, 0);
	if (!nbuf)
		goto error;

	if (GetCBuf (buf, nbuf, len) != len)
		goto error;

	obuf = buf->cb_start;
	buf->cb_start = nbuf;
	buf->cb_end = nbuf + newsize;
	buf->cb_tail = nbuf;
	buf->cb_head = nbuf + len;
	_HeapFree (obuf, 0);
	UnlockCBuf (buf);
	return newsize;
	
error:
	UnlockCBuf (buf);
	if (nbuf) _HeapFree(nbuf, 0);
	return -1;
}

void
LockCBuf (PCBUF buf)
{
	_EnterMutex (buf->cb_mutex, BLOCK_SVC_INTS);
}

int
TryLockCBuf (PCBUF buf)
{
	PTCB lockthr;
	int gotit = 0;
	_asm pushf
	_asm cli
	lockthr = _GetMutexOwner (buf->cb_mutex);
	if (!lockthr) {
		_EnterMutex (buf->cb_mutex, 0);
		gotit = 1;
	}
	_asm popf
	return gotit;
}

void
UnlockCBuf (PCBUF buf)
{
	_LeaveMutex (buf->cb_mutex);
}

int
PeekCBuf (PCBUF buf, PCHAR out, DWORD maxlen)
{
	DWORD len, partiallen;

	LockCBuf (buf);

	len = LenCBuf(buf);
	if (len > maxlen)
		len = maxlen;

	if (buf->cb_tail + len >= buf->cb_end) {
		/* wrap! */
		partiallen = buf->cb_end - buf->cb_tail;
		memcpy(out, buf->cb_tail, partiallen);
		memcpy(out+partiallen, buf->cb_start, len-partiallen);
	} else {
		memcpy(out, buf->cb_tail, len);
	}

	UnlockCBuf (buf);
	return len;
}

int
GetCBuf (PCBUF buf, PCHAR out, DWORD maxlen)
{
	DWORD len, partiallen;

	LockCBuf (buf);

	len = LenCBuf(buf);
	if (len > maxlen)
		len = maxlen;

	if (buf->cb_tail + len >= buf->cb_end) {
		/* wrap! */
		partiallen = buf->cb_end - buf->cb_tail;
		memcpy(out, buf->cb_tail, partiallen);
		memcpy(out+partiallen, buf->cb_start, len-partiallen);
		buf->cb_tail = buf->cb_start + len - partiallen;
	} else {
		memcpy(out, buf->cb_tail, len);
		buf->cb_tail += len;
	}

	UnlockCBuf (buf);
	return len;
}

void
PopCBuf (PCBUF buf, DWORD len)
{
	LockCBuf (buf);
	buf->cb_tail += len;
	if (buf->cb_tail >= buf->cb_end)
		buf->cb_tail = buf->cb_start + ((buf->cb_tail - buf->cb_start) % (buf->cb_end - buf->cb_start));
}

int
PutCBuf (PCBUF buf, PCHAR in, DWORD len)
{
	int partiallen;

	LockCBuf (buf);

	if (AvailCBuf(buf) <= (int) len) {
		UnlockCBuf (buf);
		return -1;
	}

	if (buf->cb_head + len >= buf->cb_end) {
		/* wrap! */
		partiallen = buf->cb_end - buf->cb_head;
		memcpy (buf->cb_head, in, partiallen);
		memcpy (buf->cb_start, in + partiallen, len - partiallen);
		buf->cb_head = buf->cb_start + (len - partiallen);
	} else {
		memcpy (buf->cb_head, in, len);
		buf->cb_head += len;
	}
	UnlockCBuf (buf);
	return 0;
}
